package
{
	import flash.geom.Point;
	import starling.events.Event;
	import starling.core.Starling;
	import starling.display.Sprite;
	import starling.events.Touch;
	import starling.events.TouchEvent;
	import starling.events.TouchPhase;
	import starling.extensions.ParticleDesignerPS;
	import starling.textures.Texture;
	
	/**
	 * BTTF-inspired Fire Trails
	 * Made for Sci-Fi Coding Challenge: http://active.tutsplus.com/articles/challenge/activetuts-workshop-3-sci-fi-particle-effects-coding-challenge/
	 * @author MichaelJW
	 */
	public class Demo extends Sprite
	{
		// particle data files
		// generated with http://onebyonedesign.com/flash/particleeditor/
		[Embed(source="particle.pex",mimeType="application/octet-stream")]
		private var InitValues:Class
		[Embed(source="texture.png")]
		private var Sample:Class
		
		// used to figure out whether the user is dragging a line, or
		// clicking to define start and end points. See onTouch() for more.
		private var prevPhase:String;
		private var startPoint:Point = null;
		
		public function Demo()
		{
			addEventListener(Event.ADDED_TO_STAGE, init);
		}
		
		private function init(e:Event):void
		{
			removeEventListener(Event.ADDED_TO_STAGE, init);
			
			stage.color = 0; //turn stage colour to black
			stage.addEventListener(TouchEvent.TOUCH, onTouch);
		}
		
		private function onTouch(e:TouchEvent):void
		{
			// map the touch onto stage's coordinate grid
			var touch:Touch = e.getTouch(stage);
			
			// we can ignore HOVER and STATIONARY as they aren't significant -- we don't care whether the mouse is moving
			// without being clicked, or whether the mouse is standing still
			if (touch.phase != TouchPhase.HOVER && touch.phase != TouchPhase.STATIONARY)	
			{
				// if the user clicks without dragging, we'll see a BEGAN immediately followed by an ENDED
				// (because we're ignoring HOVER and STATIONARY)
				if (prevPhase == TouchPhase.BEGAN && touch.phase == TouchPhase.ENDED)
				{
					// now we need to figure out whether this is the start point or the end point
					// we have to work backwards for this; we know it's the end point if a start point has already been set
					if (startPoint)
					{
						// we create a trail from the end point (current point) to the start point
						// seems backwards, but remember, the fire moves in that direction
						createTrail(new Point(touch.globalX, touch.globalY), startPoint);
						// we reset the start point so that we can set another one
						startPoint = null;
					}
					else
					{
						// there was no start point set, so this must be it
						startPoint = new Point(touch.globalX, touch.globalY);
					}
				}
				
				// remember that MOVED means "moved while the mouse is held down"
				// so we don't need to check BEGAN or ENDED here
				if (touch.phase == TouchPhase.MOVED)
				{
					// if the user clicked, then clicked again while dragging, then clicked a third time, we don't
					// want a line to appear from the first click to the third - so, forget the first click
					startPoint = null;
					// we create a trail from the current point to the previous point, based on where the mouse is being dragged
					createTrail(new Point(touch.globalX, touch.globalY), new Point(touch.previousGlobalX, touch.previousGlobalY));
				}
				
				// we set the previous phase so we can remember it later
				// of course, this isn't really the previous phase, but rather the previous *significant* phase
				prevPhase = touch.phase;
			}
			
		}
		
		/**
		 * Create a fire trail between two points. Works by creating a series of particle emitters, which 
		 * emit particles backwards from each point to the previous.
		 * @param	startPoint	Point where the trail starts. Fire flows from end point to start point.
		 * @param	endPoint	Point where the trail ends. Fire flows from end point to start point.
		 * @param	spread		Width of the "tyre track".
		 * @param	separation	Maximum distance between two particle emitters.
		 */
		private function createTrail(startPoint:Point, endPoint:Point, spread:Number = 20.0, separation:Number = 40):void
		{
			var newFlow:ParticleDesignerPS;
			var distance:Number = Point.distance(startPoint, endPoint);
			// angle between the start and end point, in radians
			var angle:Number = Math.atan2(endPoint.y - startPoint.y, endPoint.x - startPoint.x);
			// number of particle emitters we'll require based on the max distance between two of them
			var numParticleEmitters:int = Math.max(int(distance / separation), 1);
			var cPoint:Point;
			
			for (var i:int = 1; i <= numParticleEmitters; i++)
			{
				// the basic setup for the particles was generated using http://onebyonedesign.com/flash/particleeditor/
				newFlow = new ParticleDesignerPS(XML(new InitValues()), Texture.fromBitmap(new Sample()));
				addChild(newFlow);
				
				// we split the trail line up into equal segments based on our max distance between two emitters
				// note that because i is in [1, n], if there is only one emitter it gets placed at the *end*,
				// not the start. This makes sense, because the fire flows backwards from the emitter.
				cPoint = Point.interpolate(startPoint, endPoint, i / numParticleEmitters);
				// we move the latest emitter to the latest point along the line
				newFlow.emitterX = cPoint.x;
				newFlow.emitterY = cPoint.y;
				
				// we have to adjust the spread based on the angle, so that the width of the track is the same regardless
				// of the direction
				newFlow.emitterXVariance = spread * Math.sin(angle);
				newFlow.emitterYVariance = spread * Math.cos(angle);
				
				newFlow.emitAngle = angle;	//in radians!
				newFlow.start();
				Starling.juggler.add(newFlow);
			}
		}
	}

}